/**
 * libSpiff - XSPF playlist handling library
 *
 * Copyright (C) 2006 Sebastian Pipping
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sebastian Pipping, webmaster@hartwork.org
 */

#include "SpiffParser.h"
#include "SpiffPlaylistProps.h"
#include "SpiffTrack.h"
#include "SpiffCallback.h"
#include <stdio.h>
#include <algorithm>

namespace Spiff {



SpiffParser::SpiffParser(XML_Char * filename, SpiffCallback * callback)
		: depth(0), track(NULL), props(NULL), callback(callback) {
	// Check filename
	if (filename == NULL) {
		// TODO
		return;
	}
			
	// Check callback
	if (callback == NULL) {
		// TODO
		return;
	}

	// Create parser
	this->parser = XML_ParserCreate(NULL);

	// Put class pointer into user data
	XML_SetUserData(this->parser, this);

	// Handlers
	XML_SetElementHandler(this->parser, masterStart, masterEnd);
	XML_SetCharacterDataHandler(this->parser, masterCharacters);

	// Open file
	FILE * file = ::PORT_FOPEN(filename, _PT("r"));
	if (file == NULL) {
		return;
	}

	// Get filesize
	::fseek(file, 0, SEEK_END);
	const long filesize = ::ftell(file);
	::fseek(file, 0, SEEK_SET);

	// Read and parse file
	void * buffer;
	if (filesize > SPIFF_MAX_BLOCK_SIZE) {
		// In several blocks
		long sizeTodo = filesize;
		while (sizeTodo > 0) {
			const long blockSize = std::min<long>(sizeTodo, SPIFF_MAX_BLOCK_SIZE);
			buffer = malloc(blockSize);
			::fread(buffer, 1, blockSize, file);
			sizeTodo -= blockSize;
			if (XML_Parse(this->parser, reinterpret_cast<char *>(buffer), blockSize, sizeTodo == 0) == XML_STATUS_ERROR) {
				printf("line %i; %s", XML_GetCurrentLineNumber(this->parser),
					XML_ErrorString(XML_GetErrorCode(this->parser)));
				break;
			}
		}
	} else {
		// One single go
		buffer = XML_GetBuffer(this->parser, filesize);
		::fread(buffer, 1, filesize, file);
		if (XML_ParseBuffer(this->parser, filesize, 1) == XML_STATUS_ERROR) {
			printf("line %i; %s", XML_GetCurrentLineNumber(this->parser),
				XML_ErrorString(XML_GetErrorCode(this->parser)));
		}
	}

	// Cleanup
	::fclose(file);
	XML_ParserFree(this->parser);
}



void SpiffParser::handleStart(const XML_Char * name, const XML_Char ** atts) {
	this->depth++;

	switch (this->depth) {
	case 1:
		if (!PORT_STRCMP(name, _PT("playlist"))) {
			// TODO Check
			// xmlns = "http://xspf.org/ns/0/"

			this->stack.push(TAG_PLAYLIST);
			this->props = new SpiffPlaylistProps();
		} else {
			this->stack.push(TAG_UNKNOWN);
		}
		break;

	case 2:
		if (!PORT_STRCMP(name, _PT("title"))) {
			this->stack.push(TAG_TITLE);
		} else if (!PORT_STRCMP(name, _PT("trackList"))) {
			this->stack.push(TAG_TRACKLIST);
		} else {
			this->stack.push(TAG_UNKNOWN);
		}
		break;

	case 3:
		if (!PORT_STRCMP(name, _PT("track"))) {
			this->stack.push(TAG_TRACK);
			this->track = new SpiffTrack();
		} else {
			this->stack.push(TAG_UNKNOWN);
		}
		break;

	case 4:
		if (!PORT_STRCMP(name, _PT("location"))) {
			this->stack.push(TAG_LOCATION);
		} else if (!PORT_STRCMP(name, _PT("title"))) {
			this->stack.push(TAG_TITLE);
		} else if (!PORT_STRCMP(name, _PT("creator"))) {
			this->stack.push(TAG_CREATOR);
		} else if (!PORT_STRCMP(name, _PT("album"))) {
			this->stack.push(TAG_ALBUM);
		} else if (!PORT_STRCMP(name, _PT("trackNum"))) {
			this->stack.push(TAG_TRACKNUM);
		} else {
			this->stack.push(TAG_UNKNOWN);
		}
		break;

	default:
 		this->stack.push(TAG_UNKNOWN);
		break;

	}
/*
	// TODO remove
	// if (this->stack.top() != TAG_UNKNOWN) {
		for (int i = 1; i < this->depth; i++)
			printf("  ");

		printf(name);
		printf("\n");
	// }
*/
}



void SpiffParser::handleEnd(const XML_Char * name) {
	switch (this->depth) {
	case 1:
		if (this->stack.top() == TAG_PLAYLIST) {
			// TODO Check playlist standard conformance here?

			// Call property callback
			this->callback->setProperties(this->props);
			delete this->props;
			this->props = NULL;
		}
		break;

	case 2:
		if (this->stack.top() == TAG_TITLE) {
			this->props->setTitle(this->accum.c_str());
			this->accum.clear();
		}
		break;

	case 3:
		if (this->stack.top() == TAG_TRACK) {
			// TODO Check track standard conformance here?

			// Call property callback
			this->callback->addTrack(this->track);
			delete this->track;
			this->track = NULL;
		}
		break;

	case 4:
		switch (this->stack.top()) {
		case TAG_LOCATION:
			this->track->setLocation(this->accum.c_str());
			this->accum.clear();
			break;

		case TAG_TITLE:
			this->track->setTitle(this->accum.c_str());
			this->accum.clear();
			break;

		case TAG_CREATOR:
			this->track->setCreator(this->accum.c_str());
			this->accum.clear();
			break;

		case TAG_ALBUM:
			this->track->setAlbum(this->accum.c_str());
			this->accum.clear();
			break;

		case TAG_TRACKNUM:
			this->track->setTrackNum(this->accum.c_str());
			this->accum.clear();
			break;

		default:
			break;
		}
		break;

	default:
		break;

	}

	this->depth--;
	this->stack.pop();
}



void SpiffParser::handleCharacters(const XML_Char * s, int len) {
	switch (this->depth) {
	case 2:
		if (this->stack.top() == TAG_TITLE) {
			this->accum.append(s, len);
		}
		break;

	case 4:
		if (this->stack.top() != TAG_UNKNOWN) {
			this->accum.append(s, len);
		}
		break;

	default:
		break;

	}
}



/*static*/ inline void SpiffParser::masterStart(void * userData, const XML_Char * name, const XML_Char ** atts) {
	SpiffParser * const parser = reinterpret_cast<SpiffParser *>(userData);
	parser->handleStart(name, atts);
}



/*static*/ inline void SpiffParser::masterEnd(void * userData, const XML_Char * name) {
	SpiffParser * const parser = reinterpret_cast<SpiffParser *>(userData);
	parser->handleEnd(name);
}



/*static*/ inline void SpiffParser::masterCharacters(void * userData, const XML_Char * s, int len) {
	SpiffParser * const parser = reinterpret_cast<SpiffParser *>(userData);
	parser->handleCharacters(s, len);
}



}
